from chtoolset import query as chquery
from collections import defaultdict
from toposort import toposort
from typing import List, Tuple, Any
from tinybird.ch_utils.constants import ENABLED_TABLE_FUNCTIONS

VALID_REMOTE = 'VALID_REMOTE'


def sql_get_used_tables(sql: str, raising: bool = False, default_database: str = '',
                        table_functions: bool = True) -> List[Any]:
    """
    >>> sql_get_used_tables("SELECT 1 FROM the_table")
    [('', 'the_table', '')]
    >>> sql_get_used_tables("SELECT 1 FROM the_database.the_table")
    [('the_database', 'the_table', '')]
    >>> sql_get_used_tables("SELECT * from numbers(100)")
    [('', '', 'numbers')]
    >>> sql_get_used_tables("SELECT * FROM table1, table2")
    [('', 'table1', ''), ('', 'table2', '')]
    """
    try:
        tables: List[Tuple[str, str, str]] = chquery.tables(sql, default_database=default_database)
        if not table_functions:
            return [(t[0], t[1]) for t in tables if t[0] or t[1]]
        return tables
    except ValueError as e:
        if raising:
            raise e
        return [(default_database, sql, '')]


class ReplacementsDict(dict):
    def __getitem__(self, key):
        v = super().__getitem__(key)
        if isinstance(v, tuple):
            k, r = v
            if callable(r):
                r = r()
                super().__setitem__(key, (k, r))
            return k, r
        if callable(v):
            v = v()
            super().__setitem__(key, v)
        return v


def tables_or_sql(replacement: dict, table_functions=False) -> set:
    try:
        return set(sql_get_used_tables(replacement[1], default_database=replacement[0],
                                       raising=True, table_functions=table_functions))
    except Exception as e:
        if replacement[1][0] == '(':
            raise e
        return {replacement}


class InvalidFunction(ValueError):
    def __init__(self, function_names):
        self.msg = f"The query uses disabled table functions: '{function_names}'"
        super().__init__(self.msg)


def _separate_as_tuple_if_contains_database_and_table(definition: str) -> Any:
    if "." in definition:
        database_and_table_separated = definition.split(".")
        return database_and_table_separated[0], database_and_table_separated[1]
    return definition


def replacements_to_tuples(replacements: dict) -> dict:
    parsed_replacements = {}
    for k, v in replacements.items():
        parsed_replacements[_separate_as_tuple_if_contains_database_and_table(k)] \
            = _separate_as_tuple_if_contains_database_and_table(v)
    return parsed_replacements


class InvalidResource(ValueError):
    def __init__(self, database, table):
        self.msg = f"{database}.{table}" if database else table
        self.msg = f"Resource '{self.msg}' not found"
        super().__init__(self.msg)
        self.database = database
        self.table = table


def replace_tables(sql: str, replacements: dict, default_database: str = '', check_functions: bool = False, only_replacements: bool = False) -> str:
    """Given a query and a list of table replacements, returns the query after applying the table replacements.
    It takes into account dependencies between replacement subqueries (if any)
    It also replaces any sleep/sleepEachRow call with a call to sleep(0)/sleepEachRow(0)
    """
    if not replacements:
        # Always call replace_tables so it applies other transformations too (remove sleeps and format the query)
        return chquery.replace_tables(sql, {})

    _replaced_with = set()
    _replacements = ReplacementsDict()
    for k, r in replacements.items():
        rk = k if isinstance(k, tuple) else (default_database, k)
        _replacements[rk] = r if isinstance(r, tuple) else (default_database, r)
        _replaced_with.add(r)

    deps: defaultdict = defaultdict(set)
    _tables = sql_get_used_tables(sql, default_database=default_database, raising=True, table_functions=check_functions)
    seen_tables = set()
    while _tables:
        table = _tables.pop()
        if len(table) == 3:
            if table[2] and table[2] not in ENABLED_TABLE_FUNCTIONS:
                raise InvalidFunction(table[2])
            if table[0] and table[1]:
                table = (table[0], table[1])
            else:
                continue
        seen_tables.add(table)
        if table in _replacements:
            replacement = _replacements[table]
            dependent_tables = tables_or_sql(replacement, table_functions=check_functions)
            deps[table] |= {(d[0], d[1]) for d in dependent_tables}
            for dependent_table in list(dependent_tables):
                if len(dependent_table) == 3:
                    if dependent_table[2] and dependent_table[2] not in ENABLED_TABLE_FUNCTIONS and not(dependent_table[2] in ['remote', 'cluster'] and replacement[0] == VALID_REMOTE):
                        raise InvalidFunction(dependent_table[2])
                    if dependent_table[0] and dependent_table[1]:
                        dependent_table = (dependent_table[0], dependent_table[1])
                    else:
                        continue
                if dependent_table not in seen_tables:
                    _tables.append(dependent_table)
        else:
            deps[table] |= set()
    deps_sorted = list(reversed(list(toposort(deps))))

    if not deps_sorted:
        return chquery.replace_tables(sql, {})

    for current_deps in deps_sorted:
        current_replacements = {}
        for r in current_deps:
            if r in _replacements:
                replacement = _replacements[r]
                current_replacements[r] = replacement
                dt = chquery.table_if_is_simple_query(replacement[1], default_database=replacement[0])
                if dt:
                    current_replacements[r] = (dt[0], dt[1])
            else:
                if only_replacements:
                    continue
                database, table_name = r
                if (table_name and default_database != '' and database not in [default_database, 'tinybird', VALID_REMOTE] and r not in _replaced_with):
                    raise InvalidResource(database, table_name)

        if current_replacements:
            sql = chquery.replace_tables(sql, current_replacements, default_database=default_database)
        else:
            sql = chquery.replace_tables(sql, {})

    return sql



import json
import click
from tinybird.client import TinyB
from tinybird.tb_cli_modules.tinyunit.tinyunit_lib import DataUnitTest, customDataUnitTestDecoder, MyJSONEncoder
from os.path import exists
import requests
import glob
import urllib.parse


def test_load_file(test_file):
    existingDataList = []
    click.echo(f"Loading file {test_file}")
    if (not exists(test_file)):
        click.echo("Test file not found, creating...")
    else:
        with open(test_file) as fi:
            existingData = json.load(fi)
            for unit_data in existingData:
                unitDataTest = json.loads(unit_data, object_hook=customDataUnitTestDecoder)
                addedDataUnitTest = DataUnitTest(unitDataTest.id, unitDataTest.description, unitDataTest.enabled, unitDataTest.endpoint, unitDataTest.result, unitDataTest.time, unitDataTest.sql)
                existingDataList.append(addedDataUnitTest)

    return existingDataList


def test_write_file(test_file, inDataList):
    click.echo("Writing to file...")
    with open(test_file, 'w') as of:
        of.write(json.dumps(inDataList, cls=MyJSONEncoder, indent=4))


def test_file_add_test(tb_client: TinyB, file, endpoint, time, d, enabled, sql='', response=''):
    existingDataList = test_load_file(file)

    headers = {'Authorization': f'Bearer {tb_client.token}'}

    if(endpoint and len(endpoint) > 0):
        response = requests.get(endpoint, headers=headers)

    existingDataList.append(DataUnitTest(
        len(existingDataList),
        d,
        enabled,
        endpoint,
        response.text if endpoint else '',
        time,
        sql))

    test_write_file(file, existingDataList)
    return 0


def test_file_remove_test(file, testId):
    existingDataList = test_load_file(file)

    del existingDataList[testId]
    tmpList = []
    i = 0
    for tmp_unit_data in existingDataList:
        tmpDataUnitTest = DataUnitTest(i, tmp_unit_data.description, tmp_unit_data.enabled, tmp_unit_data.endpoint, tmp_unit_data.result, tmp_unit_data.time, tmp_unit_data.sql)
        tmpList.append(tmpDataUnitTest)
        i += 1
    existingDataList = tmpList

    test_write_file(file, existingDataList)
    return 0


def test_file_set_test_state(file, testId=None, newState=True):
    existingDataList = test_load_file(file)

    if(testId is None):
        for unitTest in existingDataList:
            unitTest.enabled = newState
    else:
        existingDataList[testId].enabled = newState

    test_write_file(file, existingDataList)
    return 0


def test_file_show_test(file, testId=None):
    existingDataList = test_load_file(file)

    if (testId is None):
        for unitTest in existingDataList:
            printDataUnitTest(unitTest)
    else:
        printDataUnitTest(existingDataList[testId])

    return 0


def test_file_reload_test(tb_client: TinyB, file, testId=None):
    existingDataList = test_load_file(file)

    headers = {'Authorization': f'Bearer {tb_client.token}'}

    if (testId is None):
        for unitTest in existingDataList:
            if(unitTest.endpoint):
                unitTest.result = requests.get(unitTest.endpoint, headers=headers).text
    else:
        existingDataList[testId].result = requests.get(existingDataList[testId].endpoint, headers=headers).text

    test_write_file(file, existingDataList)
    return 0


def printDataUnitTest(dataUnitTest):
    click.secho('Description:', fg='green', bold=True)
    click.echo(dataUnitTest.description)
    click.secho('Enabled:', fg='green', bold=True)
    click.echo(dataUnitTest.enabled)
    click.secho('Time:', fg='green', bold=True)
    click.echo(dataUnitTest.time)
    click.secho('SQL:', fg='green', bold=True)
    click.echo(dataUnitTest.sql)
    click.secho('Endpoint:', fg='green', bold=True)
    click.echo(dataUnitTest.endpoint)
    click.secho('Result:', fg='green', bold=True)
    click.echo(dataUnitTest.result)


def tinyUnitRunner(tb_client: TinyB):
    QUERY_API = f"{tb_client.host}/v0/sql?q="

    headers = {'Authorization': f'Bearer {tb_client.token}'}

    for file in glob.glob("./tests/*.json"):
        with open(glob.glob(file)[0]) as inputfile:
            data = json.load(inputfile)
            click.echo(f"->Running test from file {inputfile.name}")
            for unit_data in data:
                unitDataTest = json.loads(unit_data, object_hook=customDataUnitTestDecoder)
                if unitDataTest.enabled:
                    click.echo(f"\t->Running test: {unitDataTest.id} , {unitDataTest.description}")
                    if(unitDataTest.endpoint):
                        parsed = urllib.parse.urlparse(unitDataTest.endpoint)
                        replacedUrl = parsed._replace(netloc=getBareUrl(tb_client.host)).geturl()
                        response = requests.get(replacedUrl, headers=headers)
                        storedResponseJSON = json.loads(unitDataTest.result)
                        if response.status_code == 200:
                            click.echo("\t\t-->HTTP Response OK")
                        else:
                            click.echo("\t\t-->HTTP Response FAIL")
                            continue
                        requestedJson = json.loads(response.text)
                        if str(requestedJson["meta"]) == str(storedResponseJSON["meta"]):
                            click.echo("\t\t-->Meta Test OK")
                        else:
                            click.echo("\t\t-->Meta Test FAIL")

                        if str(requestedJson["data"]) == str(storedResponseJSON["data"]):
                            click.echo("\t\t-->Data Test OK")
                        else:
                            click.echo("\t\t-->Data Test FAIL")

                        if float(requestedJson["statistics"]["elapsed"]) * 1000 < unitDataTest.time:
                            click.echo("\t\t-->Time Test OK")
                        else:
                            click.echo("\t\t-->Time Test FAIL")
                    elif(unitDataTest.sql):
                        response = requests.get(QUERY_API + unitDataTest.sql, headers=headers)
                        if response.status_code == 200:
                            click.echo("\t\t-->HTTP Response OK")
                        else:
                            click.echo("\t\t-->HTTP Response FAIL")
                            continue
                        if(len(response.text) == 0):
                            click.echo("\t\t-->SQL Test OK")
                        else:
                            click.echo("\t\t-->SQL Test FAIL")


def getBareUrl(url):
    if url.startswith("http://"):
        return url[7:]
    elif url.startswith("https://"):
        return url[8:]
    else:
        return url
